{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# FastGlobalAvgPool"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "加载所需包："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import testing\n",
    "import numpy as np\n",
    "import tvm\n",
    "from tvm import relay\n",
    "from tvm.relay.dataflow_pattern import (\n",
    "    wildcard, is_op, \n",
    "    is_constant, \n",
    "    is_tuple,\n",
    "    # is_tuple_get_item,\n",
    "    DFPatternCallback,\n",
    "    rewrite\n",
    ")\n",
    "from tvm.relay import transform as _transform"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "构建模型："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "from torch import nn\n",
    "from torch.nn import functional as F\n",
    "\n",
    "class FastGlobalAvgPool(nn.Module):\n",
    "    def __init__(self, flatten=False):\n",
    "        super().__init__()\n",
    "        self.flatten = flatten\n",
    "\n",
    "    def forward(self, x):\n",
    "        if self.flatten:\n",
    "            in_size = x.size()\n",
    "            return x.view((in_size[0], in_size[1], -1)).mean(dim=2)\n",
    "        else:\n",
    "            return x.view(x.size(0), x.size(1), -1).mean(-1).view(x.size(0), x.size(1), 1, 1)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "模型推理："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "xx = torch.randn(1, 64, 8, 9)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "tensor([ 0.0662, -0.0284, -0.0418, -0.0225,  0.0498,  0.1133, -0.2982,  0.1316,\n",
       "         0.0266,  0.0706, -0.2277,  0.0217,  0.1399, -0.1460, -0.0213,  0.0273,\n",
       "        -0.1333,  0.0970, -0.0018,  0.0033, -0.1182,  0.1267, -0.2266,  0.1154,\n",
       "        -0.2021,  0.0986, -0.0339,  0.0458, -0.0835, -0.2597, -0.1705, -0.0668,\n",
       "         0.0603,  0.0841, -0.0064, -0.0208,  0.0534,  0.1446,  0.0612,  0.0570,\n",
       "         0.0104, -0.1448, -0.0248,  0.0502, -0.0378, -0.0554, -0.0469,  0.1171,\n",
       "        -0.0959, -0.0214, -0.1768,  0.0065, -0.0934, -0.0269, -0.1720,  0.1292,\n",
       "        -0.0320,  0.0377,  0.0316, -0.2030,  0.0909, -0.2396, -0.0421, -0.0082])"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "m1 = FastGlobalAvgPool()\n",
    "y1 = m1(xx)\n",
    "y1.squeeze()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "等价于："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "tensor([ 0.0662, -0.0284, -0.0418, -0.0225,  0.0498,  0.1133, -0.2982,  0.1316,\n",
       "         0.0266,  0.0706, -0.2277,  0.0217,  0.1399, -0.1460, -0.0213,  0.0273,\n",
       "        -0.1333,  0.0970, -0.0018,  0.0033, -0.1182,  0.1267, -0.2266,  0.1154,\n",
       "        -0.2021,  0.0986, -0.0339,  0.0458, -0.0835, -0.2597, -0.1705, -0.0668,\n",
       "         0.0603,  0.0841, -0.0064, -0.0208,  0.0534,  0.1446,  0.0612,  0.0570,\n",
       "         0.0104, -0.1448, -0.0248,  0.0502, -0.0378, -0.0554, -0.0469,  0.1171,\n",
       "        -0.0959, -0.0214, -0.1768,  0.0065, -0.0934, -0.0269, -0.1720,  0.1292,\n",
       "        -0.0320,  0.0377,  0.0316, -0.2030,  0.0909, -0.2396, -0.0421, -0.0082])"
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "m2 = nn.AdaptiveAvgPool2d(1)\n",
    "y2 = m2(xx)\n",
    "y2.squeeze()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "数值一致性检验："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "np.testing.assert_allclose(y1.numpy(), y2.numpy())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "转换为 Relay 模型："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "name = \"x\"\n",
    "shape = (1, 1024, 8, 6)\n",
    "data_np = (np.random.randint(0, 256, shape)/255).astype(\"float32\")\n",
    "data_torch = torch.from_numpy(data_np)\n",
    "\n",
    "model = FastGlobalAvgPool().eval()\n",
    "scripted_model = torch.jit.trace(model, data_torch).eval()\n",
    "shape_list = [(name, shape)]\n",
    "mod, params = relay.frontend.from_pytorch(scripted_model, shape_list)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div class=\"highlight\" style=\"background: \"><pre style=\"line-height: 125%;\"><span></span><span style=\"color: #008000; font-weight: bold\">def</span> <span style=\"color: #AA22FF\">@main</span>(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x: Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::size_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>x:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>) <span style=\"color: #AA22FF; font-weight: bold\">-&gt;</span> Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] {\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">48</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::view_0:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> mean(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, axis<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::mean_0:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::view_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>\n",
       "}\n",
       "</pre></div>\n"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "_transform.InferType()(tvm.IRModule.from_expr(mod[\"main\"])).show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "\u001b[0;31mInit signature:\u001b[0m \u001b[0mFastGlobalAvgPoolSimplify\u001b[0m\u001b[0;34m(\u001b[0m\u001b[0;34m)\u001b[0m\u001b[0;34m\u001b[0m\u001b[0;34m\u001b[0m\u001b[0m\n",
      "\u001b[0;31mDocstring:\u001b[0m     \n",
      "简化 reshape+mean+reshape 为 nn.adaptive_avg_pool2d(%x, output_size=[1]) \n",
      "\n",
      "简化 \n",
      "    def @main(%x: Tensor[(1, 1024, 8, 6), float32] /* ty=Tensor[(1, 1024, 8, 6), float32] span=aten::size_0.x:0:0 */) -> Tensor[(1, 1024, 1, 1), float32] {\n",
      "        %0 = reshape(%x, newshape=[1, 1024, -1]) /* ty=Tensor[(1, 1024, 48), float32] span=aten::view_0:0:0 */;\n",
      "        %1 = mean(%0, axis=[-1]) /* ty=Tensor[(1, 1024), float32] span=aten::mean_0:0:0 */;\n",
      "        reshape(%1, newshape=[1, 1024, 1, 1]) /* ty=Tensor[(1, 1024, 1, 1), float32] span=aten::view_1:0:0 */\n",
      "        }\n",
      "为\n",
      "    def @main(%x: Tensor[(1, 1024, 8, 6), float32] /* ty=Tensor[(1, 1024, 8, 6), float32] span=aten::size_0.x:0:0 */) {\n",
      "        nn.adaptive_avg_pool2d(%x, output_size=[1]) /* ty=Tensor[(1, 1024, 1, 1), float32] */\n",
      "        }\n",
      "\u001b[0;31mFile:\u001b[0m           /media/pc/data/lxw/ai/tvm-book/src/tvm_book/transforms/simplify.py\n",
      "\u001b[0;31mType:\u001b[0m           type\n",
      "\u001b[0;31mSubclasses:\u001b[0m     "
     ]
    }
   ],
   "source": [
    "from tvm_book.transforms.simplify import FastGlobalAvgPoolSimplify\n",
    "\n",
    "FastGlobalAvgPoolSimplify?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "run_mod = tvm.IRModule()\n",
    "run_mod[\"main\"] = rewrite(FastGlobalAvgPoolSimplify(), mod[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div class=\"highlight\" style=\"background: \"><pre style=\"line-height: 125%;\"><span></span><span style=\"color: #008000; font-weight: bold\">def</span> <span style=\"color: #AA22FF\">@main</span>(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x: Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::size_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>x:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>) {\n",
       "  nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>adaptive_avg_pool2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x, output_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">*/</span>\n",
       "}\n",
       "</pre></div>\n"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "run_mod.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "验证数值一致性："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [],
   "source": [
    "with tvm.transform.PassContext(opt_level=3, disabled_pass={\"AlterOpLayout\"}):\n",
    "    lib = relay.build(mod, target=\"llvm\", params=params)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(tvm.cpu(0)))\n",
    "module.run(**{name: data_np})\n",
    "num_outputs = module.get_num_outputs()\n",
    "origin_outputs = [module.get_output(k).numpy() for k in range(num_outputs)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "with tvm.transform.PassContext(opt_level=3, disabled_pass={\"AlterOpLayout\"}):\n",
    "    lib = relay.build(run_mod, target=\"llvm\", params=params)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(tvm.cpu(0)))\n",
    "module.run(**{name: data_np})\n",
    "num_outputs = module.get_num_outputs()\n",
    "outputs = [module.get_output(k).numpy() for k in range(num_outputs)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [],
   "source": [
    "np.testing.assert_allclose(origin_outputs[0], outputs[0])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "ai",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
